<?php
/* vim: set noexpandtab tabstop=2 softtabstop=2 shiftwidth=2: */

/**
 * TMXInfoFetcher - Fetch TMX info/records for TMO/TMS/TMN/TMU(F)/TMNF tracks
 * Created by Xymph <tm@gamers.org> based on https://api2.mania-exchange.com/
 * Inspired by TMNDataFetcher & "Stats for TMN Aseco+RASP"
 *
 * v2.0: Updated to use the current ManiaExchange API; added $authtime, $authscor,
 *       $vehicle; removed deprecated $visible, $length, $lbrating, $game;
 *       $thumburl is no longer a small version of $imageurl but the GBX
 *       thumbnail, if available; in $recordlist, added 'position', 'rscore',
 *       'respawns' entries, removed 'approved', 'expires', 'lockspan' entries
 * v1.21: Added https support to get_file
 * v1.20: Fixed constructor for PHP 8.x, updated URLs to https
 * v1.19: Allowed 24-char UIDs too
 * v1.18: Added 'replayurl' to the entries in $recordslist
 * v1.17: Allowed 25-char UIDs too
 * v1.16: Improved error reporting via $error; parsed more comment formatting
 * v1.15: Added User-Agent to the GET request
 * v1.14: Fixed PHP Strict level warning
 * v1.13: Fixed handling of empty API responses
 * v1.12: Renamed $worldrec into $custimg to correct its meaning; fixed
 *        $replayid check
 * v1.11: Added another check for valid $replayid
 * v1.10: Added magic __set_state function to support var_export()
 * v1.9: Fixed fetch records run-time warnings
 * v1.8: Optimized get_file URL parsing
 * v1.7: Added $worldrec (boolean); changed $visible to boolean; renamed
 *       $download to $dloadurl, $imgurl to $imageurl & $imgurlsmall to
 *       $thumburl; minor tweaks
 * v1.6: Added check for valid $replayid & $replayurl
 * v1.5: Added TMNF compatibility
 * v1.4: Added get_file function to handle TMX site timeouts
 * v1.3: Added $awards, $comments, $replayid, $replayurl; renamed $comment to
 *       $acomment (to better distinguish author comment from # of comments)
 * v1.2: Allowed 26-char UIDs too; added $pageurl
 * v1.1: Allowed TMX IDs too
 * v1.0: Initial release
 */
class TMXInfoFetcher {

	public $section, $site, $uid, $id, $records, $error,
		$name, $userid, $author, $authtime, $authscor, $uploaded, $updated,
		$type, $envir, $mood, $style, $routes, $diffic, $vehicle,
		$awards, $comments, $custimg, $acomment, $pageurl, $replayid,
		$replayurl, $imageurl, $thumburl, $dloadurl, $recordlist;

	/**
	 * Fetches a hell of a lot of data about a TMX track
	 *
	 * @param String $game
	 *        TMX section for 'TMO', 'TMS', 'TMN', 'TMU', 'TMNF'
	 * @param String $id
	 *        The challenge UID to search for (if a 24-27 char alphanum string),
	 *        otherwise the TMX ID to search for (if a number)
	 * @param Boolean $records
	 *        If true, the script also returns the world records (max. 10)
	 * @return TMXInfoFetcher
	 *        If $error is not an empty string, it's an error message
	 */
	public function __construct($game, $id, $records) {

		$this->section = $game;
		switch ($game) {
		case 'TMO':
			$this->site = 'original.tm-exchange.com';
			break;
		case 'TMS':
			$this->site = 'sunrise.tm-exchange.com';
			break;
		case 'TMN':
			$this->site = 'nations.tm-exchange.com';
			break;
		case 'TMU':
			$this->site = 'tmuf.exchange';
			break;
		case 'TMNF':
			$this->site = 'tmnf.exchange';
			break;
		default:
			$this->site = '';
			return;
		}

		$this->error = '';
		$this->records = $records;
		// check for UID string
		if (preg_match('/^\w{24,27}$/', $id)) {
			$this->uid = $id;
			$this->getData(true);
		// check for TMX ID
		} elseif (is_numeric($id) && $id > 0) {
			$this->id = intval($id);
			$this->getData(false);
		}
	}  // __construct

	public static function __set_state($import) {

		$tmx = new TMXInfoFetcher('', 0, false);

		$tmx->section   = $import['section'];
		if (isset($import['site']))
			$tmx->site    = $import['site'];
		else
			switch($tmx->section) {
			case 'TMO':
				$tmx->site  = 'original.tm-exchange.com';
				break;
			case 'TMS':
				$tmx->site  = 'sunrise.tm-exchange.com';
				break;
			case 'TMN':
				$tmx->site  = 'nations.tm-exchange.com';
				break;
			case 'TMU':
				$tmx->site  = 'tmuf.exchange';
				break;
			case 'TMNF':
				$tmx->site  = 'tmnf.exchange';
				break;
			}

		$tmx->uid       = $import['uid'];
		$tmx->id        = (int)$import['id'];
		$tmx->records   = $import['records'];
		$tmx->error     = '';
		$tmx->name      = $import['name'];
		$tmx->userid    = (int)$import['userid'];
		$tmx->author    = $import['author'];
		$tmx->authtime  = isset($import['authtime']) ? $import['authtime'] : 0;
		$tmx->authscor  = isset($import['authscor']) ? $import['authscor'] : 0;
		$tmx->uploaded  = $import['uploaded'];
		$tmx->updated   = $import['updated'];
		$tmx->type      = $import['type'];
		$tmx->envir     = $import['envir'];
		$tmx->mood      = $import['mood'];
		$tmx->style     = $import['style'];
		$tmx->routes    = $import['routes'];
		$tmx->diffic    = $import['diffic'];
		$tmx->vehicle   = isset($import['vehicle']) ? $import['vehicle'] : '';
		$tmx->awards    = (int)$import['awards'];
		$tmx->comments  = (int)$import['comments'];
		$tmx->custimg   = $import['custimg'];
		$tmx->acomment  = $import['acomment'];
		$tmx->pageurl   = $import['pageurl'];
		$tmx->replayid  = $import['replayid'];
		$tmx->replayurl = $import['replayurl'];
		$tmx->imageurl  = $import['imageurl'];
		$tmx->thumburl  = $import['thumburl'];
		$tmx->dloadurl  = $import['dloadurl'];
		$tmx->recordlist = null;

		return $tmx;
	}  // __set_state

	private function getData($isuid) {

		// map enums to strings
		$envirs = array(
			1 => 'Snow',
			2 => 'Desert',
			3 => 'Rally',
			4 => 'Island',
			5 => 'Coast',
			6 => 'Bay',
			7 => 'Stadium',
		);
		$cars = array(
			1 => 'SnowCar',
			2 => 'DesertCar',
			3 => 'RallyCar',
			4 => 'IslandCar',
			5 => 'CoastCar',
			6 => 'BayCar',
			7 => 'StadiumCar',
		);
		$moods = array(
			0 => 'Sunrise',
			1 => 'Day',
			2 => 'Sunset',
			3 => 'Night',
		);
		$types = array(
			0 => 'Race',
			1 => 'Puzzle',
			2 => 'Platform',
			3 => 'Stunts',
			4 => 'Shortcut',
			5 => 'Laps',
		);
		$routes = array(
			0 => 'Single',
			1 => 'Multiple',
			2 => 'Symmetrical',
		);
		$diffcts = array(
			0 => 'Beginner',
			1 => 'Intermediate',
			2 => 'Expert',
			3 => 'Lunatic',
		);
		$tags = array(
			0 => 'Normal',
			1 => 'Stunt',
			2 => 'Maze',
			3 => 'Offroad',
			4 => 'Laps',
			5 => 'Fullspeed',
			6 => 'LOL',
			7 => 'Tech',
			8 => 'SpeedTech',
			9 => 'RPG',
			10 => 'PressForward',
			11 => 'Trial',
			12 => 'Grass',
		);

		// define output parameters
		$fields = array(
			'TrackId',
			'TrackName',
			'UId',
			'Uploader.UserId',
			'Uploader.Name',
			'AuthorTime',
			'AuthorScore',
			'PrimaryType',
			'Environment',
			'Mood',
			'Style',
			'Routes',
			'Difficulty',
			'Car',
			'Awards',
			'Comments',
			'HasThumbnail',
			'WRReplay.ReplayId',
			'UploadedAt',
			'UpdatedAt',
			'AuthorComments',
			'Images',
		);

		// get track info
		$url = "https://{$this->site}/api/tracks?";
		if ($isuid)
			$url .= "uid={$this->uid}";
		else
			$url .= "id={$this->id}";
		$file = $this->get_file($url . "&fields=" .
			rawurlencode(implode(',', $fields))
		);
		if ($file === false) {
			$this->error = 'Connection or response error on ' . $url;
			return;
		} else if ($file === -1) {
			$this->error = 'Timed out while reading data from ' . $url;
			return;
		} else if ($file == '') {
			$this->error = 'No data returned from ' . $url;
			return;
		}

		// process map info
		$tmx = json_decode($file);
		if ($tmx === null) {
			$this->error = 'Cannot decode JSON data from ' . $url;
			return;
		}
		if (!isset($tmx->Results) || empty($tmx->Results)) {
			$this->error = 'No data returned from ' . $url;
			return;
		}
		$tmx = $tmx->Results[0];

		if ($isuid)
			$this->id     = $tmx->TrackId;
		else
			$this->uid    = $tmx->UId;
		$this->name     = $tmx->TrackName;
		$this->userid   = $tmx->Uploader->UserId;
		$this->author   = $tmx->Uploader->Name;
		$this->authtime = $tmx->AuthorTime;
		$this->authscor = $tmx->AuthorScore;
		$this->uploaded = $tmx->UploadedAt;
		$this->updated  = $tmx->UpdatedAt;
		$this->type     = $types[ $tmx->PrimaryType ];
		$this->envir    = $envirs[ $tmx->Environment ];
		$this->mood     = $moods[ $tmx->Mood ];
		$this->style    = $tags[ $tmx->Style ];
		$this->routes   = $routes[ $tmx->Routes ];
		$this->diffic   = $diffcts[ $tmx->Difficulty ];
		$this->vehicle  = $cars[ $tmx->Car ];
		$this->awards   = $tmx->Awards;
		$this->comments = $tmx->Comments;
		$this->custimg  = !empty($tmx->Images);
		$this->replayid = isset($tmx->WRReplay->ReplayId) ? $tmx->WRReplay->ReplayId : 0;

		$search = array(chr(31), '[b]', '[/b]', '[i]', '[/i]', '[u]', '[/u]', '[url]', '[/url]');
		$replace = array('<br/>',  '<b>', '</b>', '<i>', '</i>', '<u>', '</u>', '<i>', '</i>');
		$this->acomment = str_ireplace($search, $replace, (string) $tmx->AuthorComments);
		$this->acomment = preg_replace('/\[url=".*"\]/', '<i>', $this->acomment);

		$this->pageurl  = "https://{$this->site}/trackshow/{$this->id}";
		$this->imageurl = $this->custimg ? "https://{$this->site}/trackshow/{$this->id}/image/1" : '';
		$this->thumburl = $tmx->HasThumbnail ? "https://{$this->site}/trackshow/{$this->id}/image/0" : '';
		$this->dloadurl = "https://{$this->site}/trackgbx/{$this->id}";

		if ($this->replayid > 0)
			$this->replayurl = "https://{$this->site}/recordgbx/{$this->replayid}";
		else
			$this->replayurl = '';

		// fetch records too?
		$this->recordlist = array();
		if ($this->records) {
			// define output parameters
			$fields = array(
				'ReplayId',
				'Position',
				'User.UserId',
				'User.Name',
				'ReplayTime',
				'ReplayScore',
				'ReplayRespawns',
				'ReplayAt',
				'TrackAt',
				'Score',
			);

			// get replays info (max. 10 entries)
			$url = "https://{$this->site}/api/replays?best=1&trackId={$this->id}";
			$file = $this->get_file($url . "&fields=" .
				rawurlencode(implode(',', $fields))
			);
			if ($file === false || $file === -1 || $file == '')
				// no error message as main info was already fetched
				return;

			// process replays info
			$tmx = json_decode($file);
			if ($tmx === null) {
				$this->error = 'Cannot decode JSON data from ' . $url;
				return;
			}
			$tmx = $tmx->Results;

			foreach ($tmx as $replay) {
				$this->recordlist[] = array(
					'replayid'  => $replay->ReplayId,
					'position'  => $replay->Position,
					'userid'    => $replay->User->UserId,
					'name'      => $replay->User->Name,
					'time'      => $replay->ReplayTime,
					'rscore'    => $replay->ReplayScore,
					'respawns'  => $replay->ReplayRespawns,
					'replayat'  => $replay->ReplayAt,
					'trackat'   => $replay->TrackAt,
					'score'     => (int)$replay->Score,
					'replayurl' => "https://{$this->site}/recordgbx/{$replay->ReplayId}",
				);
			}
		}
	}  // getData

	// Simple HTTP(S) Get function with timeout
	// ok: return string || error: return false || timeout: return -1
	private function get_file($url) {

		$url = parse_url($url);
		$port = isset($url['port']) ? $url['port'] :
		        ($url['scheme'] == 'https' ? 443 : 80);
		$query = isset($url['query']) ? '?' . $url['query'] : '';

		$fp = @fsockopen(($url['scheme'] == 'https' ? 'ssl://' : '') . $url['host'],
		                 $port, $errno, $errstr, 4);
		if (!$fp)
			return false;

		fwrite($fp, 'GET ' . $url['path'] . $query . " HTTP/1.0\r\n" .
		            'Host: ' . $url['host'] . "\r\n" .
		            'Content-Type: application/json' . "\r\n" .
		            'User-Agent: TMXInfoFetcher (' . PHP_OS . ")\r\n\r\n");
		stream_set_timeout($fp, 4);
		$res = '';
		$info = array();
		$info['timed_out'] = false;
		while (!feof($fp) && !$info['timed_out']) {
			$res .= fread($fp, 512);
			$info = stream_get_meta_data($fp);
		}
		fclose($fp);

		if ($info['timed_out']) {
			return -1;
		} else {
			if (substr($res, 9, 3) != '200')
				return false;
			$page = explode("\r\n\r\n", $res, 2);
			return trim($page[1]);
		}
	}  // get_file
}  // class TMXInfoFetcher

